#!/usr/bin/python

import bzrlib.tests
import os.path

class ShelfTests(bzrlib.tests.TestCaseWithTransport):
    ORIGINAL = '\n\nhello test world\n\n'
    MODIFIED = '\n\ngoodbye test world\n\n'
    DIFF_HEADER = "=== modified file 'a/test_file'\n"
    DIFF_1 = """--- a/test_file\t
+++ b/test_file\t
@@ -1,4 +1,4 @@
 
 
-hello test world
+goodbye test world
 
"""
    DIFF_2 = """--- a/test_file\t
+++ b/test_file\t
@@ -1,4 +1,4 @@
 
 
-goodbye test world
+hello test world
 
"""
    def test_shelf(self):
        self.__test_loop(1)

    def test_shelf_multi(self):
        self.__test_loop(10)

    def __test_loop(self, count):
        tree = self.make_branch_and_tree('.')
        self.__create_and_add_test_file(tree)

        while count > 0:
            count -= 1

            # Modify the test file
            file('test_file', 'w').write(self.MODIFIED)

            # Check the diff is right
            self.assertEqual(self.capture('diff', retcode=1),
                self.DIFF_HEADER + self.DIFF_1 + '\n')

            # Shelve the changes
            self.run_bzr('shelve', '--all', retcode=0)

            # Make sure there is no diff anymore
            self.assertEqual(self.capture('diff', retcode=0), '')

            # Make sure the file is actually back the way it was
            self.assertEqual(file('test_file').read(), self.ORIGINAL)

            # Check the shelf is right
            shelf = open(os.path.join(tree.branch.base,
                        '.shelf/shelves/default/00')).read()
            shelf = shelf[shelf.index('\n') + 1:] # skip the message
            self.assertEqual(shelf, self.DIFF_1)

            # Unshelve
            self.run_bzr('unshelve', '--all', retcode=0)

            # Check the diff is right again
            self.assertEqual(self.capture('diff', retcode=1),
                self.DIFF_HEADER + self.DIFF_1 + '\n')

            # Check the shelved patch was backed up
            shelf = open(os.path.join(tree.branch.base,
                        '.shelf/shelves/default/00~')).read()
            shelf = shelf[shelf.index('\n') + 1:] # skip the message
            self.assertEqual(shelf, self.DIFF_1)

            # Make sure the file is back the way it should be
            self.assertEqual(file('test_file').read(), self.MODIFIED)

    def test_shelf_nothing_to_shelve(self):
        import os.path
        tree = self.make_branch_and_tree('.')
        self.__create_and_add_test_file(tree)

        # Shelve the changes
        self.run_bzr('shelve', '--all', retcode=3)

        if os.path.exists(os.path.join(tree.branch.base,
                '.shelf/shelves/default/00')):
            self.fail("Shelf exists, but it shouldn't")

    def __create_and_add_test_file(self, tree, filename='test_file'):
        f = open(filename, 'w')
        f.write(self.ORIGINAL)
        f.close()
        tree.add(tree.relpath(os.path.join(os.getcwd(), filename)))
        tree.commit(message='add %s' % filename)

    def test_shelf_with_revision(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)

        # Modify the test file and commit it
        self.build_tree_contents([('test_file', self.MODIFIED)])
        tree.commit(message='update test_file')

        # Shelve the changes
        self.run_bzr('shelve', '--all', '-r', '1', retcode=0)

        # Check the diff is right
        self.assertEqual(self.capture('diff', retcode=1),
            self.DIFF_HEADER + self.DIFF_2 + '\n')

        # Make sure the file is the way it should be
        self.assertEqual(file('test_file').read(), self.ORIGINAL)

        # Unshelve
        self.run_bzr('unshelve', '--all', retcode=0)

        # Make sure the file is back the way it should be
        self.assertEqual(file('test_file').read(), self.MODIFIED)

    def test_shelf_with_two_revisions(self):
        tree = self.make_branch_and_tree('.')

        cmd = 'shelve --all -r 1..2'
        (stdout, stderr) = self.run_bzr_captured(cmd.split(), retcode=None)

        self.assertEqual(stderr.split('\n')[0],
            'bzr: ERROR: shelve only accepts a single revision parameter.')

    def test_shelf_show_basic(self):
        tree = self.make_branch_and_tree('.')
        self.__create_and_add_test_file(tree)
        self.__test_show(tree, '00')

    def __test_show(self, tree, patch):
        # Modify the test file
        self.build_tree_contents([('test_file', 'patch %s\n' % patch)])

        # Shelve the changes
        self.run_bzr('shelve', '--all', retcode=0)

        # Make sure there is no diff anymore
        self.assertEqual(self.capture('diff', retcode=0), '')

        # Check the shelf is right
        shelf = open(os.path.join(tree.branch.base,
                    '.shelf/shelves/default', patch)).read()
        self.assertTrue('patch %s' % patch in shelf)

        # Check the shown output is right
        shown = self.capture('shelf show %s' % patch, retcode=0)
        self.assertEqual(shown, shelf)

    def test_shelf_show_multi(self):
        tree = self.make_branch_and_tree('.')
        self.__create_and_add_test_file(tree)
        self.__test_show(tree, '00')
        self.__test_show(tree, '01')
        self.__test_show(tree, '02')

        # Now check we can show a previously shelved patch
        shelf = open(os.path.join(tree.branch.base,
                    '.shelf/shelves/default/00')).read()
        self.assertTrue('patch 00' in shelf)

        # Check the shown output is right
        shown = self.capture('shelf show 00', retcode=0)
        self.assertEqual(shown, shelf)

    def test_shelf_show_with_no_patch(self):
        tree = self.make_branch_and_tree('.')
        stderr = self.run_bzr_captured(['shelf', 'show', '00'], retcode=None)[1]
        self.assertTrue("Patch '00' doesn't exist on shelf default!" in stderr)

    def test_shelf_unshelve_failure(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)

        # Modify the test file
        file('test_file', 'w').write(self.MODIFIED)

        # Shelve the changes
        self.run_bzr('shelve', '--all', retcode=0)

        # Write an unapplyable patch into the shelf
        shelf = open(os.path.join(tree.branch.base,
                    '.shelf/shelves/default/00'), 'w')
        shelf.write(self.DIFF_2)
        shelf.close()

        # Unshelve, should fail
        self.run_bzr('unshelve', '--all', retcode=3)

        # Make sure the patch is still there, eventhough it's broken
        shelf = open(os.path.join(tree.branch.base,
                    '.shelf/shelves/default/00')).read()
        self.assertEqual(shelf, self.DIFF_2)

        # Working tree should be unchanged
        diff = self.capture('diff', retcode=0)
        self.assertEqual(diff, '')

    def test_shelf_unshelve_failure_two_hunks(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)
        self.__create_and_add_test_file(tree, filename='test_file2')

        # Modify the test files
        file('test_file', 'w').write(self.MODIFIED)
        file('test_file2', 'w').write(self.MODIFIED)

        # Shelve the changes
        self.run_bzr('shelve', '--all', retcode=0)

        # Put the changes to test_file back, the shelved patch won't apply now
        file('test_file', 'w').write(self.MODIFIED)
        tree.commit(message='screw up test_file')

        # Unshelve, should fail
        self.run_bzr('unshelve', '--all', retcode=3)

        # Working tree should be unchanged
        diff = self.capture('diff', retcode=0)
        self.assertEqual(diff, '')

        # Force should succeed and modify test_file2, but leave shelf
        self.run_bzr('unshelve', '--force', '--all', retcode=0)
        self.assertEqual(open('test_file2').read(), self.MODIFIED)
        self.assertTrue(os.path.exists('.shelf/shelves/default/00'))

    def test_shelf_after_unshelve(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)
        self.__create_and_add_test_file(tree, filename='test_file2')

        # Modify the test files
        file('test_file', 'w').write(self.MODIFIED)
        file('test_file2', 'w').write(self.MODIFIED)

        # Shelve the changes
        self.run_bzr('shelve', '--all', 'test_file', retcode=0)
        self.run_bzr('shelve', '--all', 'test_file2', retcode=0)

        # Unshelve
        self.run_bzr('unshelve', '--all', retcode=0)

        # We should now have 00 and 01~
        self.assertTrue(os.path.exists('.shelf/shelves/default/00'))
        self.assertTrue(os.path.exists('.shelf/shelves/default/01~'))

        # Check ls works
        list = self.capture('shelf ls', retcode=0).split('\n')
        for line in list:
            self.assertFalse(line.startswith(' 01'))

        # Unshelve, if unshelve is confused by the backup it will fail
        self.run_bzr('unshelve', '--all', retcode=0)

    def test_shelf_delete(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)
        self.__create_and_add_test_file(tree, filename='test_file2')

        # Modify the test files
        file('test_file', 'w').write(self.MODIFIED)
        file('test_file2', 'w').write(self.MODIFIED)

        # Shelve the changes
        self.run_bzr('shelve', '--all', 'test_file', retcode=0)
        self.run_bzr('shelve', '--all', 'test_file2', retcode=0)

        # Delete 00
        self.run_bzr('shelf', 'delete', '00', retcode=0)

        # We should now have 01 but not 00
        self.assertFalse(os.path.exists('.shelf/shelves/default/00'))
        self.assertTrue(os.path.exists('.shelf/shelves/default/01'))

        # Check ls works
        list = self.capture('shelf ls', retcode=0).split('\n')
        for line in list:
            self.assertFalse(line.startswith(' 00'))

        # Unshelve should unshelve 01
        self.run_bzr('unshelve', '--all', retcode=0)
        self.assertEqual(file('test_file2').read(), self.MODIFIED)

    def test_shelf_gaps(self):
        tree = self.make_branch_and_tree('.')
        self.__create_and_add_test_file(tree)
        file('test_file', 'w').write(self.MODIFIED)
        self.run_bzr('shelve', '--all', 'test_file', retcode=0)
        file('test_file', 'w').write(self.MODIFIED)
        self.run_bzr('shelve', '--all', 'test_file', retcode=0)

        # Now delete 00, leaving 01, next shelve should go into 02
        self.run_bzr('shelf', 'delete', '0', retcode=0)
        self.assertFalse(os.path.exists('.shelf/shelves/default/00'))
        self.assertFalse(os.path.exists('.shelf/shelves/default/02'))
        file('test_file', 'w').write(self.MODIFIED)
        self.run_bzr('shelve', '--all', 'test_file', retcode=0)
        self.assertFalse(os.path.exists('.shelf/shelves/default/00'))
        self.assertTrue(os.path.exists('.shelf/shelves/default/02'))

    def test_shelf_upgrade(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)

        # Modify then shelve, so we're not upgrading to 00, just for kicks
        file('test_file', 'w').write(self.MODIFIED)
        self.run_bzr('shelve', '--all', 'test_file', retcode=0)

        open('.bzr-shelf', 'w').write('First old shelf')
        open('.bzr-shelf-1', 'w').write('Second old shelf')
        open('.bzr-shelf-3', 'w').write('Fourth old shelf')

        # shelve and unshelve should bitch and do nothing
        file('test_file', 'w').write('blah blah blah')
        self.run_bzr('shelve', '--all', retcode=3)
        self.assertFalse(os.path.exists('.shelf/shelves/default/01'))
        self.assertEqual(file('test_file').read(), 'blah blah blah')
        self.run_bzr('unshelve', '--all', retcode=3)
        self.assertTrue(os.path.exists('.shelf/shelves/default/00'))

        # Upgrade, make sure it worked
        self.run_bzr('shelf', 'upgrade', retcode=0)
        self.assertEqual(open('.shelf/shelves/default/01').read(),
                'First old shelf')
        self.assertEqual(open('.shelf/shelves/default/02').read(),
                'Second old shelf')
        self.assertEqual(open('.shelf/shelves/default/03').read(),
                'Fourth old shelf')

        # Check the old shelves got backed up right
        self.assertTrue(os.path.exists('.bzr-shelf~'))
        self.assertTrue(os.path.exists('.bzr-shelf-1~'))
        self.assertTrue(os.path.exists('.bzr-shelf-3~'))
        self.assertFalse(os.path.exists('.bzr-shelf'))
        self.assertFalse(os.path.exists('.bzr-shelf-1'))
        self.assertFalse(os.path.exists('.bzr-shelf-3'))

        # Shelve should work now
        self.run_bzr('shelve', '--all', retcode=0)

    def test_shelf_p0_patch(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)

        # Run a benign shelf command to setup .shelf for us
        self.run_bzr('shelf', 'ls', retcode=0)

        # Fake a -p0 shelved patch
        diff = self.DIFF_1
        diff = diff.replace('a/', '')
        diff = diff.replace('b/', '')
        open('.shelf/shelves/default/00', 'w').write(diff)

        # This should work
        self.run_bzr('unshelve', '--all', retcode=0)

        # Check the diff is right
        self.assertEqual(self.capture('diff', retcode=1),
            self.DIFF_HEADER + self.DIFF_1 + '\n')

    def test_shelf_shelve_in_subdir(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)

        # Modify the test file
        file('test_file', 'w').write(self.MODIFIED)

        # Shelve the changes
        self.run_bzr('shelve', '--all', retcode=0)

        # Working tree should be unchanged
        diff = self.capture('diff', retcode=0)
        self.assertEqual(diff, '')

        # Unshelve, should succeed
        self.run_bzr('unshelve', '--all', retcode=0)

        # Check the diff is right
        self.assertEqual(self.capture('diff', retcode=1),
            self.DIFF_HEADER + self.DIFF_1 + '\n')
